'''
/***************************************************************************
Name	   :  region_controller.py
Description:  Region controller class for FEWSTools plugin,
              updated from QGIS2
copyright  :  (C) 2019-2022 by FEWS
email      :  cholen@contractor.usgs.gov
Modified   :  08/14/2020 cholen - Fix regex check on comments, use utils
              08/29/2020 cholen - Add mask import functionality
              10/02/2020 cholen - Fix default index select
              02/23/2022 cholen - Codebase consistency changes.
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
'''
import re
import os

from PyQt5.QtWidgets import QMessageBox, QDialog, QFileDialog

from . import fews_tools_config as config
from . import geoclim_utilities as util

from .new_name_controller import NewNameController
from .get_spatial_extents_controller import GetSpatialExtentsController
from .import_vector_to_raster_controller import ImportVectorToRasterController
from .region_model import RegionModel
from .workspace_setup_model import WorkspaceSetupModel
from .Ui_DefineRegion import Ui_DefineRegion


class RegionController(QDialog):
    '''
    Class for region definitions
    '''
    def __init__(self):
        QDialog.__init__(self)
        self.ui = Ui_DefineRegion()
        self.ui.setupUi(self)
        self.region_name_list = None

        # get a list to hold widget values
        self.widget_val_list = [''] * 11

        self.reg_info = RegionModel()
        self.wrksp_setup = WorkspaceSetupModel()  # need to get paths

        self.map_files_loc = self.wrksp_setup.get_map_data_path()
        self.static_files_loc = self.wrksp_setup.get_static_data_path()

        # will track changes to form for 'save'
        self.ui.commentLineEdit.textChanged.connect(self.save_enabled)
        self.ui.defaultRegComboBox.currentIndexChanged['QString'].connect(
            self.save_enabled)
        self.ui.minLatSpinBox.valueChanged.connect(self.save_enabled)
        self.ui.maxLatSpinBox.valueChanged.connect(self.save_enabled)
        self.ui.minLonSpinBox.valueChanged.connect(self.save_enabled)
        self.ui.maxLonSpinBox.valueChanged.connect(self.save_enabled)
        self.ui.cellSizeSpinBox.valueChanged.connect(self.save_enabled)
        self.ui.maskLineEdit.textChanged.connect(self.save_enabled)
        self.ui.mapLineEdit.textChanged.connect(self.save_enabled)
        self.ui.map2LineEdit.textChanged.connect(self.save_enabled)

        # connect remaining widgets
        self.ui.regionComboBox.currentIndexChanged['QString'].connect(
            self.update_form_widgets_for_new_region_selection)
        self.ui.getExtentFromMapButton.clicked.connect(
            self.get_extent_from_map)
        self.ui.browseMapButton.clicked.connect(self.browse_map1_path)
        self.ui.browseMap2Button.clicked.connect(self.browse_map2_path)
        self.ui.browseMaskButton.clicked.connect(self.browse_mask_path)
        self.ui.saveButton.clicked.connect(self.save_region)
        self.ui.deleteButton.clicked.connect(self.delete_region)
        self.ui.newButton.clicked.connect(self.new_region)
        self.ui.copyRegionButton.clicked.connect(self.new_region_from_copy)
        self.ui.importButton.clicked.connect(self.import_region)
        self.ui.importMapButton.clicked.connect(self.import_mask_from_shape)

        self.__initial_form_load__()

    def __check_new_name__(self, new_name):
        '''
        Check if name exists return err if yes
        :param new_name(string) - New name
        :return err(boolean) - True if name already exists, else False
        '''
        err = False
        self.region_name_list = self.reg_info.get_all_region_names()
        if new_name in self.region_name_list or \
                self.ui.regionComboBox.findText(new_name) != -1:
            err = True
            QMessageBox.critical(
                self, 'Error - Name already exists',
                'The selected name already exists in the database.',
                QMessageBox.Ok)
        return err

    def __check_required_widgets__(self):
        '''
        Check that required widgets are filled.
        :return err(boolean) - True if needed entry is missing, else False
        '''
        form_err = False
        req_combo_boxes = [self.ui.regionComboBox.currentText(),
                           self.ui.defaultRegComboBox.currentText()]
        req_line_edits = [self.ui.maskLineEdit.text(),
                          self.ui.mapLineEdit.text()]
        # verify that all required fields are filled:
        for entry in req_combo_boxes:
            if entry in (' ', ''):
                form_err = True
        for entry in req_line_edits:
            if entry == '':
                form_err = True
        return form_err

    def __initial_form_load__(self):
        '''
        Fills the form entries initial load
        '''
        util.fill_misc_widget(self.ui.defaultRegComboBox, ['No', 'Yes'])

        self.region_name_list = self.reg_info.get_all_region_names()
        self.region_name_list = sorted(self.region_name_list)
        util.fill_misc_widget(self.ui.regionComboBox, self.region_name_list)
        # get name for default region from db
        default_reg_name = self.reg_info.query_default_region_name()
        # changing the index
        # forces update_form_widgets_for_new_region_selection which fills form
        self.ui.regionComboBox.setCurrentIndex(
            self.ui.regionComboBox.findText(default_reg_name))

        self.ui.saveButton.setEnabled(False)

    def __validate_text_inputs__(self):
        '''
        Validate the free form text entries
        return err(boolean) - True if entry has an error, else False
        '''
        err = False
        checks = self.ui.commentLineEdit.text()
        # comments can be empty
        # but must not contain special chars or be too long
        for entry in checks:
            if len(entry) > config.MAX_CHARS or not\
                    re.match(config.REG_EXPR_COMMENTS, entry):
                err = True
        if err is True:
            QMessageBox.critical(
                self,
                'Error in form comments',
                'Comments may contain alpha-numeric characters,\n' +
                'underscores and dashes.\n' +
                'No other special characters are allowed!',
                QMessageBox.Ok)
        return err

    def browse_map_path(self, widget):
        '''
        Function to browse to map files.
        :param widget(LineEdit widget object) - The path widget
        '''
        map_file =\
            QFileDialog.getOpenFileName(self,
                                        'Please Select the Map File',
                                        self.map_files_loc,
                                        config.VEC_FILE_TYPES)
        if map_file[0] != '':
            widget.setText(os.path.basename(map_file[0]))

    def browse_map1_path(self):
        '''
        Function to browse to map files.
        '''
        self.browse_map_path(self.ui.mapLineEdit)

    def browse_map2_path(self):
        '''
        Function to browse to alternate map files.
        '''
        self.browse_map_path(self.ui.map2LineEdit)

    def browse_mask_path(self):
        '''
        Function to browse to mask files.
        '''
        mask_file =\
            QFileDialog.getOpenFileName(self,
                                        'Please Select the Mask File',
                                        self.static_files_loc,
                                        config.RST_FILE_TYPES)
        if mask_file[0] != '':
            self.ui.maskLineEdit.setText(os.path.basename(mask_file[0]))

    def delete_region(self):
        '''
        Delete selected region from the database.
        '''
        reg_name = self.ui.regionComboBox.currentText()
        reply = QMessageBox.question(
            self,
            'Delete Region?',
            'Are you sure you want to delete the ' + reg_name + ' region?\n '
            'This cannot be undone.',
            QMessageBox.Yes, QMessageBox.Cancel)

        if reply == QMessageBox.Yes:
            self.reg_info.delete_region_record(reg_name)
            # need to fix the combo box
            self.ui.regionComboBox.removeItem(
                self.ui.regionComboBox.findText(reg_name))
            # reload the form
            show_region = self.reg_info.query_default_region_name()
            self.ui.regionComboBox.setCurrentIndex(
                self.ui.regionComboBox.findText(show_region))
            if reg_name in self.region_name_list:
                self.region_name_list.remove(reg_name)

    def get_extent_from_map(self):
        '''
        Fill in map extents from a selected map.
        '''
        extents_dlg = GetSpatialExtentsController()
        extents_dlg.exec_()
        extents = extents_dlg.get_extents()
        self.ui.minLatSpinBox.setValue(extents[0])
        self.ui.maxLatSpinBox.setValue(extents[1])
        self.ui.minLonSpinBox.setValue(extents[2])
        self.ui.maxLonSpinBox.setValue(extents[3])

    def get_widget_val_list(self):
        '''
        Get the widget values list
        '''
        return self.widget_val_list

    def import_mask_from_shape(self):
        '''
        Initiates the dialog to rasterize a vector into a mask file
        '''
        region_fields_dic = {
            'RegionName': self.ui.regionComboBox.currentText(),
            'MinimumLatitude': self.ui.minLatSpinBox.value(),
            'MaximumLatitude': self.ui.maxLatSpinBox.value(),
            'MinimumLongitude': self.ui.minLonSpinBox.value(),
            'MaximumLongitude': self.ui.maxLonSpinBox.value(),
            'Height': self.ui.cellSizeSpinBox.value(),
            'Width': self.ui.cellSizeSpinBox.value()}
        import_mask_dlg = ImportVectorToRasterController(region_fields_dic)
        import_mask_dlg.exec_()
        # need to assign the result to the mask control
        self.ui.maskLineEdit.setText(os.path.basename(
            import_mask_dlg.get_mask_filename()))

    def import_region(self):
        '''
        Function to import a region from a .region file.
        '''
        region_file =\
            QFileDialog.getOpenFileName(
                self,
                'Please Select a Region File',
                self.wrksp_setup.get_climate_data_path(),
                'Region files (*.region);;All files (*.*)')

        if region_file[0]:
            # the name of the region will be the basename of the file(no ext)
            new_reg_name = os.path.basename(region_file[0]).split('.region')[0]
            err = self.__check_new_name__(new_reg_name)
            if err is False:
                reply = QMessageBox.question(
                    self,
                    'New default region?',
                    'Do you want to set this as the default region?',
                    QMessageBox.Yes, QMessageBox.No)

                self.ui.regionComboBox.addItem(new_reg_name)
                self.ui.regionComboBox.setCurrentIndex(
                    self.ui.regionComboBox.count() - 1)
                if reply == QMessageBox.Yes:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('Yes'))
                else:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('No'))
                self.ui.commentLineEdit.clear()
                self.ui.mapLineEdit.clear()
                self.ui.map2LineEdit.clear()
                # parse the text file to get the values....
                with open(region_file[0], 'r') as reg_text:
                    for line in reg_text:
                        if ',' in line:
                            val = line.split(',')[1]
                            if 'minLat' in line:
                                self.ui.minLatSpinBox.setValue(float(val))
                            elif 'maxLat' in line:
                                self.ui.maxLatSpinBox.setValue(float(val))
                            elif 'minLon' in line:
                                self.ui.minLonSpinBox.setValue(float(val))
                            elif 'maxLon' in line:
                                self.ui.maxLonSpinBox.setValue(float(val))
                            elif 'PixelLat' in line:
                                self.ui.cellSizeSpinBox.setValue(float(val))
                            elif 'PixelLon' in line:
                                self.ui.cellSizeSpinBox.setValue(float(val))
                            # NOTE - these files must exist in the workspace
                            elif 'MaskName' in line:
                                filename = os.path.basename(val)
                                self.ui.maskLineEdit.setText(filename)
                            # there are two map names allowed
                            elif 'defaultMapFile' in line:
                                filename = os.path.basename(val)
                                if self.ui.mapLineEdit.text() in (' ', ''):
                                    self.ui.mapLineEdit.setText(filename)
                                else:
                                    self.ui.map2LineEdit.setText(filename)
                self.save_region()

    def new_region(self):
        '''
        Function to create a new region from a reset form.
        '''
        new_region_name_dlg = NewNameController('Region')
        new_region_name_dlg.exec_()
        new_reg_name = new_region_name_dlg.get_new_name()

        if new_reg_name:
            err = self.__check_new_name__(new_reg_name)
            if err is False:
                reply = QMessageBox.question(
                    self,
                    'New default region?',
                    'Do you want to set this as the default region?',
                    QMessageBox.Yes, QMessageBox.No)

                self.ui.regionComboBox.addItem(new_reg_name)
                self.ui.regionComboBox.setCurrentIndex(
                    self.ui.regionComboBox.count() - 1)
                if reply == QMessageBox.Yes:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('Yes'))
                else:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('No'))
                self.ui.minLatSpinBox.setValue(0.000)
                self.ui.maxLatSpinBox.setValue(0.000)
                self.ui.minLonSpinBox.setValue(0.000)
                self.ui.maxLonSpinBox.setValue(0.000)
                self.ui.cellSizeSpinBox.setValue(0.050)
                self.ui.commentLineEdit.clear()
                self.ui.mapLineEdit.clear()
                self.ui.maskLineEdit.clear()

    def new_region_from_copy(self):
        '''
        Create a new region by copying an existing region.
        '''
        new_region_name_dlg = NewNameController('Region')
        new_region_name_dlg.exec_()
        new_reg_name = new_region_name_dlg.get_new_name()
        if new_reg_name:
            err = self.__check_new_name__(new_reg_name)
            if err is False:
                reply = QMessageBox.question(
                    self, 'New default region?',
                    'Do you want to set this as the default region?',
                    QMessageBox.Yes, QMessageBox.No)

                self.ui.regionComboBox.addItem(new_reg_name)
                self.ui.regionComboBox.setCurrentIndex(
                    self.ui.regionComboBox.count() - 1)
                if reply == QMessageBox.Yes:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('Yes'))
                else:
                    self.ui.defaultRegComboBox.setCurrentIndex(
                        self.ui.defaultRegComboBox.findText('No'))
                self.save_region()

    def save_region(self):
        '''
        Update region database using widget selections.
        '''
        w_err = self.__check_required_widgets__()
        t_err = self.__validate_text_inputs__()
        if w_err is False and t_err is False:
            self.set_widget_val_list()
            self.reg_info.write_region_record(self.widget_val_list)
            self.ui.saveButton.setEnabled(False)
        elif not t_err:
            QMessageBox.information(
                self, 'Error - Missing information',
                'Required fields are missing.\n'
                'Recheck form entries and try again or cancel.',
                QMessageBox.Ok)

    def save_enabled(self):
        '''
        Enable the save button. To save unnecessary writes to the
        database, the button should only be enabled if something has changed
        on the form.
        '''
        self.ui.saveButton.setEnabled(True)

    def set_widget_val_list(self):
        '''
        Get the widget values into a list
        '''
        if self.ui.map2LineEdit.text():
            map_widget_string = (self.ui.mapLineEdit.text() + ',' +
                                 self.ui.map2LineEdit.text())
        else:
            map_widget_string = self.ui.mapLineEdit.text()
        self.widget_val_list =\
            [self.ui.regionComboBox.currentText(),
             self.ui.minLatSpinBox.value(),
             self.ui.maxLatSpinBox.value(),
             self.ui.minLonSpinBox.value(),
             self.ui.maxLonSpinBox.value(),
             self.ui.cellSizeSpinBox.value(),
             self.ui.cellSizeSpinBox.value(),
             self.ui.maskLineEdit.text(),
             map_widget_string,
             self.ui.defaultRegComboBox.currentText(),
             self.ui.commentLineEdit.text()]

    def update_form_widgets_for_new_region_selection(self):
        '''
        Function to update form widgets when region selection changes.
        '''
        # use the name to set region object data structure vals
        self.reg_info.query_named_region(self.ui.regionComboBox.currentText())
        # get the values for use here
        reg_fields_dic = self.reg_info.get_region_dictionary()
        self.ui.map2LineEdit.clear()  # in case it doesn't exist in selected
        self.ui.commentLineEdit.setText(reg_fields_dic['Comments'])
        self.ui.minLatSpinBox.setValue(reg_fields_dic['MinimumLatitude'])
        self.ui.maxLatSpinBox.setValue(reg_fields_dic['MaximumLatitude'])
        self.ui.minLonSpinBox.setValue(reg_fields_dic['MinimumLongitude'])
        self.ui.maxLonSpinBox.setValue(reg_fields_dic['MaximumLongitude'])
        self.ui.cellSizeSpinBox.setValue(reg_fields_dic['Height'])
        self.ui.cellSizeSpinBox.setValue(reg_fields_dic['Width'])
        self.ui.maskLineEdit.setText(reg_fields_dic['Mask'])
        map_file_tuple = reg_fields_dic['Map'].split(',')
        self.ui.mapLineEdit.setText(map_file_tuple[0])
        if len(map_file_tuple) > 1:
            self.ui.map2LineEdit.setText(map_file_tuple[1])
        self.ui.defaultRegComboBox.setCurrentIndex(
            self.ui.defaultRegComboBox.findText(reg_fields_dic['DefaultReg']))
        self.ui.saveButton.setEnabled(False)
