'''
/***************************************************************************
Name	   :  region_model.py
Description:  Region model class for FEWSTools plugin,
              updated from QGIS2
copyright  :  (C) 2019-2022 by FEWS
email      :  cholen@contractor.usgs.gov
Modified   :  11/12/2019 - cholen - unset_default_flag_on_non_selected_region
              02/20/2020 - cholen - Add write from dictionary.
              01/20/2022 - cholen - Remove commented out code.
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
'''
import copy

from . import database_utilities


class RegionModel:
    '''
    Class to hold region information.
    '''
    def __init__(self):
        self.regions_table_name = 'Regions'
        self.region_fields_dic = None
        initial_default_region = self.query_default_region_name()
        self.query_named_region(initial_default_region)

    def delete_region_record(self, reg_name):
        '''
        Deletes selected region
        :param ds_name(string): Name of region to delete
        '''
        sql = ('DELETE FROM ' + self.regions_table_name +
               ' WHERE RegionName = ?')
        params = (reg_name, )
        database_utilities.write_db(sql, params)

    def get_all_region_names(self):
        '''
        Extract region names from regions table. Used to fill
        region combo box when loading form.
        '''
        sql = 'SELECT * FROM ' + self.regions_table_name
        qry_result = database_utilities.query_db(sql, None, True)
        region_list = []
        for row in qry_result:
            region_list.append(row[0])
        return region_list

    def get_region_dictionary(self):
        '''
        Return region dictionary
        '''
        return self.region_fields_dic

    def query_named_region(self, reg_name):
        '''
        Extracts region information from database and loads it into the class
        variables.
        :param reg_name(string): Name of region of interest.
        :return err(boolean): True if not found, else false
        '''
        err = True
        sql = ("SELECT * FROM " + self.regions_table_name +
               " WHERE RegionName=?")
        params = (str(reg_name), )
        qry_result = database_utilities.query_db(sql, params, False)
        if qry_result:  # update the dictionary with the returned query
            self.region_fields_dic = {
                'RegionName': qry_result[0],
                'MinimumLatitude': qry_result[1],
                'MaximumLatitude': qry_result[2],
                'MinimumLongitude': qry_result[3],
                'MaximumLongitude': qry_result[4],
                'Height': qry_result[5],
                'Width': qry_result[6],
                'Mask': qry_result[7],
                'Map': qry_result[8],
                'DefaultReg': qry_result[9],
                'Comments': qry_result[10]}
            err = False
        return err

    def query_default_region_name(self):
        '''
        Extracts default region name and returns it
        If none are marked as default it returns the first region record
        :return ret_val(string): Default region name
        '''
        sql = ("SELECT * FROM " + self.regions_table_name +
               " WHERE DefaultReg = 'Yes'")
        qry_result = database_utilities.query_db(sql)
        if qry_result:
            ret_val = qry_result[0]
        else:
            # get the first region in the list
            reg_name_list = self.get_all_region_names()
            ret_val = reg_name_list[0]
        return ret_val

    def unset_default_flag_on_non_selected_regions(
            self, default_reg_name):
        '''
        Update the default region.
        :params default_reg_name(string): The new default region name.
        '''
        params = (default_reg_name,)
        sql_write =\
            ("UPDATE " + self.regions_table_name +
             " SET DefaultReg = 'No' WHERE RegionName != ?")
        database_utilities.write_db(sql_write, params)

        sql_write =\
            ("UPDATE " + self.regions_table_name +
             " SET DefaultReg = 'Yes' WHERE RegionName = ?")
        database_utilities.write_db(sql_write, params)

    def write_region_record(self, widget_vals):
        '''
        Writes or updates a row in the regions table.
        :params
            widget_vals(list): The widget values from form in this order.
                 RegionName,MinimumLatitude,MaximumLatitude,
                 MinimumLongitude, MaximumLongitude,Height, Width,
                 Mask,Map, DefaultReg,Comments
        '''
        err = self.query_named_region(widget_vals[0])
        loc_widget_vals = copy.deepcopy(widget_vals)
        if err is False:
            loc_widget_vals.append(loc_widget_vals.pop(0))
            sql_write =\
                ("UPDATE " + self.regions_table_name +
                 " SET MinimumLatitude=?,MaximumLatitude=?, "
                 "MinimumLongitude=?,MaximumLongitude=?,Height=?,Width=?,"
                 "Mask=?,Map=?,DefaultReg=?,Comments=?"
                 "WHERE RegionName = ?")
        else:
            sql_write =\
                ("INSERT INTO " + self.regions_table_name +
                 " (RegionName,MinimumLatitude,MaximumLatitude,"
                 " MinimumLongitude, MaximumLongitude, Height,Width,"
                 " Mask,Map,DefaultReg,Comments)"
                 " VALUES(?,?,?,?,?,?,?,?,?,?,?)")
        params = tuple(loc_widget_vals)
        database_utilities.write_db(sql_write, params)
        # update the default region if necessary
        if widget_vals[-2] == 'Yes':
            self.unset_default_flag_on_non_selected_regions(widget_vals[0])

    def write_region_record_from_dic(self, reg_dic):
        '''
        Writes or updates a row in the regions table using a region dictionary.
        :params
           reg_dic: The 'widget' values need to be this order.
                 RegionName,MinimumLatitude,MaximumLatitude,
                 MinimumLongitude, MaximumLongitude,Height, Width,
                 Mask,Map, DefaultReg,Comments
        '''
        param_list = [
            reg_dic['RegionName'],
            reg_dic['MinimumLatitude'],
            reg_dic['MaximumLatitude'],
            reg_dic['MinimumLongitude'],
            reg_dic['MaximumLongitude'],
            reg_dic['Height'],
            reg_dic['Width'],
            reg_dic['Mask'],
            reg_dic['Map'],
            reg_dic['DefaultReg'],
            reg_dic['Comments']]
        self.write_region_record(param_list)
