'''
/***************************************************************************
Name	      :  workspace_setup_controller.py
Description:  Workspace Settings for FEWSTools plugin, updated from QGIS2
copyright  :  (C) 2019-2022 by FEWS
email      :  cholen@contractor.usgs.gov
Modified   :  02/07/2020 - CHOLEN - Add fix_os_sep_in_path to path builders
              05/19/2020 msun - Adjusted default workspace creation
              06/02/2020 - CHOLEN - Stopped copy into a folder that already has
                                    a GeoCLIM folder in place.
              07/14/2020 - CHOLEN - Log on exceptions
              02/22/2021 - cholen - Add call for initialize_default_output_dir
              02/23/2022 - cholen - Codebase consistency changes
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
'''
import getpass
import os
import platform
import shutil
import zipfile

from PyQt5 import QtCore
from PyQt5.QtWidgets import QMessageBox, QDialog, QFileDialog
from qgis.core import QgsMessageLog, Qgis

from . import fews_tools_config

from .workspace_setup_model import WorkspaceSetupModel
from .datasets_model import DatasetsModel
from .Ui_WorkspaceSettings import Ui_WorkspaceSettings


class WorkspaceSetupController(QDialog):
    '''
    Class to handle the workspace settings for FEWSTools plugin
    '''

    def __init__(self):
        QDialog.__init__(self)
        self.ui = Ui_WorkspaceSettings()
        self.ui.setupUi(self)
        self.ui.wrkSpaceLineEdit.setEnabled(False)
        self.ui.browseButton.clicked.connect(self.browse_workspace)
        self.ui.cancelButton.clicked.connect(self.cancel_browse)
        self.ui.okButton.clicked.connect(self.update_db_and_move_files)
        self.wrksp_setup = WorkspaceSetupModel()
        self.ds_info = DatasetsModel()
        self.move = False
        self.new_workspace = None
        self.old_workspace = self.wrksp_setup.get_workspace()

        # force user to select a workspace if none exists
        if not self.old_workspace:
            self.browse_workspace()

        self.ui.existWrkspace.setText(self.old_workspace)

    def move_workspace_files(self):
        '''
        Move the workspace.
        '''
        try:
            shutil.move(self.old_workspace, self.new_workspace)
        except BaseException:
            QgsMessageLog.logMessage(
                'Exception - GeoCLIM Workspace move failed',
                level=Qgis.Critical)

    def build_default_workspace(self):
        '''
        Create a default workspace. This extracts the zip provided with the
        plugin and places it in 'C:/Users/<user>/Documents'.
        '''
        default_zip = self.wrksp_setup.fix_os_sep_in_path(os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "DefaultWorkspaceTemplate.zip"))
        default_unzip = self.wrksp_setup.fix_os_sep_in_path(
            os.path.join(os.path.dirname(os.path.realpath(__file__))))
        src = self.wrksp_setup.fix_os_sep_in_path(
            os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         "DefaultWorkspaceTemplate",
                         "GeoCLIM"))
        # TODO - Complete os independence here, needs testing
        if 'Windows' in platform.platform():
            default_workspace = self.wrksp_setup.fix_os_sep_in_path(
                os.path.join('C:' + os.sep, 'Users',
                             getpass.getuser(),
                             'Documents', 'GeoCLIM'))
        else:
            # Linux and MacOS
            default_workspace = self.wrksp_setup.fix_os_sep_in_path(
                os.path.join('/bin/usr', 'GeoCLIM'))

        if os.path.exists(default_workspace):
            QMessageBox.information(
                self,
                'GeoCLIM workspace already exists.\n',
                'The GeoCLIM workspace - \n' + default_workspace +
                '\n exists. \nIt will be used as the default workspace.',
                QMessageBox.Ok)
            self.wrksp_setup.set_workspace(default_workspace)
            self.ds_info.update_dataset_filepaths(default_workspace)
        else:
            QMessageBox.information(self,
                                    'No existing GeoCLIM workspace',
                                    'Creating default GeoCLIM workspace\n' +
                                    default_workspace,
                                    QMessageBox.Ok)
            try:
                # unzip the default workspace
                zip_ref = zipfile.ZipFile(default_zip, 'r')
                zip_ref.extractall(default_unzip)
                zip_ref.close()
                shutil.copytree(src, default_workspace)
                # update the db entries to point to the default workspace
                self.wrksp_setup.set_workspace(default_workspace)
                self.ds_info.update_dataset_filepaths(default_workspace)
                # set initial default output dir
                self.wrksp_setup.initialize_default_output_dir(
                    default_workspace)
            except BaseException:
                QgsMessageLog.logMessage(
                    'Exception - GeoCLIM Workspace creation failed',
                    level=Qgis.Critical)

    def browse_workspace(self):
        '''
        Browse to desired location of workspace and updates the database value.
        '''
        folder =\
            QFileDialog.getExistingDirectory(None,
                                             "Select GeoCLIM Workspace",
                                             QtCore.QDir.currentPath())
        new_path = os.path.join(folder, fews_tools_config.GEOCLIM)
        # check for a GeoCLIM folder in selected location
        # if exists then give dialog and return, we will not allow a folder
        # that already contains a 'GeoCLIM' folder, because the copy used
        # will actually create a nested 'GeoCLIM' folder within the existing
        # 'GeoCLIM' folder
        if os.path.exists(new_path):
            QMessageBox.warning(
                self,
                "GeoCLIM Workspace Move Not Allowed",
                ("There is already a GeoCLIM folder at that location.\n"
                 "The workspace cannot be moved there.\n"
                 "Please select a valid location."),
                QMessageBox.Ok)
        elif os.path.exists(folder):
            self.new_workspace = self.wrksp_setup.fix_os_sep_in_path(new_path)
            self.ui.wrkSpaceLineEdit.setText(self.new_workspace)

    def cancel_browse(self):
        '''
        Cancel - reset new workspace to none.
        '''
        self.new_workspace = None

    def update_db_and_move_files(self):
        '''
        Update the database tables and move the files if necessary
        to the location of self.new_workspace.
        '''
        try:
            if self.ui.wrkSpaceLineEdit.text() != '':
                # don't allow a copy if the workspace is already there
                if self.old_workspace == self.new_workspace:
                    QMessageBox.warning(
                        self,
                        "GeoCLIM Workspace Move Not Allowed",
                        ("The new workspace location is the same as the "
                         "current workspace location.\n"
                         "The workspace will not be moved."),
                        QMessageBox.Ok)
                else:
                    reply = QMessageBox.question(
                        self,
                        "GeoCLIM Workspace Move Confirmation",
                        ("1) Before moving a workspace it is recommended "
                         "to create a backup.\n\n"
                         "2) Be aware that moving a large workspace can "
                         "take significant time.\n"
                         "A message box will confirm when the move "
                         "has completed.\n"
                         "QGIS and the GeoCLIM plugin will be locked "
                         "until the move is complete.\n\n"
                         "3) Moving workspace from:\n   " +
                         self.old_workspace + "\n   to:\n   " +
                         self.new_workspace +
                         ".\n\nDo you want to continue?"),
                        QMessageBox.Yes,
                        QMessageBox.Cancel)
                    if reply == QMessageBox.Yes:
                        # move the workspace files
                        self.move_workspace_files()
                        # update the worksapce
                        self.wrksp_setup.set_workspace(self.new_workspace)
                        # update the ds paths
                        self.ds_info.update_dataset_filepaths(
                            self.new_workspace)
                        QMessageBox.information(
                            self,
                            "Move GeoCLIM Workspace Succeeded",
                            "The workspace was successfully moved.",
                            QMessageBox.Ok)
        except BaseException:
            # on any error put the db settings back to what they were
            self.wrksp_setup.set_workspace(self.old_workspace)
            self.ds_info.update_dataset_filepaths(self.old_workspace)
            QgsMessageLog.logMessage(
                'Exception - Updating GeoCLIM workspace values failed',
                level=Qgis.Critical)
        else:
            # no error occurs when updating db, then update output directory
            # this function will check to see if the output dir is within the
            # workspace, only updates if it is.
            if self.new_workspace:
                self.wrksp_setup.set_output_dir(
                    self.old_workspace, self.new_workspace)
